from ._flight_offers import FlightOffers
from ._flight_dates import FlightDates
from ._flight_destinations import FlightDestinations
from ._hotel_offers import HotelOffers

__all__ = ['FlightOffers', 'FlightDates', 'FlightDestinations', 'HotelOffers']



from amadeus.namespaces._reference_data import ReferenceData
from amadeus.namespaces._travel import Travel
from amadeus.namespaces._shopping import Shopping


class Core(object):
    def __init__(self):
        self.reference_data = ReferenceData(self)
        self.travel = Travel(self)
        self.shopping = Shopping(self)



from amadeus.client.decorator import Decorator


class FlightDates(Decorator, object):
    def get(self, **params):
        '''
        Find the cheapest flight dates from an origin to a destination.

        .. code-block:: python

            amadeus.shopping.flight_dates.get(origin='LHR', destination='PAR')

        :param origin: the City/Airport IATA code from which the flight will
            depart. ``"BOS"``, for example for Boston.

        :param destination: the City/Airport IATA code to which the flight is
            going. ``"BOS"``, for example for Boston.

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: if the request could not be completed
        '''
        return self.client.get('/v1/shopping/flight-dates', **params)



import time


class AccessToken(object):
    # The number of seconds before the token expires, when
    # we will already try to refresh it
    TOKEN_BUFFER = 10

    def __init__(self, client):
        self.access_token = None
        self.expires_at = 0
        self.client = client

    # PROTECTED

    # The bearer token that can be used directly in API request headers
    def _bearer_token(self):
        return 'Bearer {0}'.format(self.__token())

    # PRIVATE

    # Returns the access token if it is still valid,
    # or refreshes it if it is not (or about to expire)
    def __token(self):
        if self.__needs_refresh():
            self.__update_access_token()
        return self.access_token

    # Checks if the token needs a refesh by checking if the token
    # is nil or (about to) expire(d)
    def __needs_refresh(self):
        has_access_token = self.access_token is not None
        current_time_window = int(time.time()) + self.TOKEN_BUFFER
        has_valid_token = current_time_window < self.expires_at

        return not (has_access_token and has_valid_token)

    # Fetches a new access token and stores it and its expiry date
    def __update_access_token(self):
        response = self.__fetch_access_token()
        self.__store_access_token(response.result)

    # Fetches a new access token
    def __fetch_access_token(self):
        return self.client._unauthenticated_request(
            'POST',
            '/v1/security/oauth2/token',
            {
                'grant_type': 'client_credentials',
                'client_id': self.client.client_id,
                'client_secret': self.client.client_secret
            }
        )

    # Store an access token and calculates the expiry date
    def __store_access_token(self, data):
        self.access_token = data.get('access_token', None)
        current_time = int(time.time())
        self.expires_at = current_time + data.get('expires_in', 0)



from amadeus.client.decorator import Decorator
from .analytics._air_traffic import AirTraffic
from .analytics._fare_searches import FareSearches


class Analytics(Decorator, object):
    def __init__(self, client):
        Decorator.__init__(self, client)
        self.fare_searches = FareSearches(client)
        self.air_traffic = AirTraffic(client)



from amadeus.client.decorator import Decorator


class FareSearches(Decorator, object):
    def get(self, **params):
        '''
        The Fare Search History API allows to find the number of
        estimated searches from an origin, optionally a destination,
        within a time period when travelers are performing the searches.

        .. code-block:: python

            amadeus.travel.analytics.fare_searches.get(
                origin='LHR',
                sourceCountry='FR'
                period='2011'
            )

        :param cityCode: IATA code of the origin city, for
            example ``"BOS"`` for Boston.
        :param sourceCountry: IATA code of the country from which fare
            searches were made - e.g. ``"US"`` for United States
        :param period: period of search; can be a year
            or a month or a week. ISO format must be used - e.g. ``"2015"``
            for year; ``"2015-05"`` for month and ``"2015-W04"`` for week.
            Period ranges are not supported. Only periods from year
            ``"2011-01"`` up to current year and previous month or week
            are valid. Future dates are not supported.

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: if the request could not be completed
        '''
        return self.client.get('/v1/travel/analytics/fare-searches', **params)



from amadeus.client.decorator import Decorator


class Offer(Decorator, object):
    def __init__(self, client, hotel_id, offer_id):
        Decorator.__init__(self, client)
        self.hotel_id = hotel_id
        self.offer_id = offer_id

    def get(self, **params):
        '''
        Get room and rate details for a specific hotel offer

        .. code-block:: python

            amadeus.shopping.hotel(
                'SMPARCOL'
            ).offers('AC7D4DA2C322A73AF0824318A4965DA2805A3FC2').get()

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: if the request could not be completed
        '''
        return self.client.get(
            '/v1/shopping/hotels/{0}/offers/{1}'.format(
                self.hotel_id, self.offer_id
            ), **params)



from amadeus.client.decorator import Decorator
from .air_traffic._traveled import Traveled
from .air_traffic._booked import Booked
from .air_traffic._busiest_period import BusiestPeriod


class AirTraffic(Decorator, object):
    def __init__(self, client):
        Decorator.__init__(self, client)
        self.booked = Booked(client)
        self.traveled = Traveled(client)
        self.busiest_period = BusiestPeriod(client)



from amadeus.client.decorator import Decorator


class FlightDestinations(Decorator, object):
    def get(self, **params):
        '''
        Find the cheapest destinations where you can fly to.

        .. code-block:: python

            amadeus.shopping.flight_destinations.get(origin='LHR')

        :param origin: the City/Airport IATA code from which the flight will
            depart. ``"BOS"``, for example for Boston.

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: if the request could not be completed
        '''

        return self.client.get('/v1/shopping/flight-destinations', **params)



from amadeus.client.decorator import Decorator


class FlightOffers(Decorator, object):
    def get(self, **params):
        '''
        Find the cheapest bookable flights.

        .. code-block:: python

            amadeus.shopping.flight_destinations.get(
                origin='LHR',
                destination='PAR',
                departureDate='2017-12-24'
            )

        :param origin: the City/Airport IATA code from which the flight will
            depart. ``"BOS"``, for example for Boston.

        :param destination: the City/Airport IATA code to which the flight is
            going. ``"BOS"``, for example for Boston.

        :param departureDate: the date on which to fly out, in `YYYY-MM-DD`
            format

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: if the request could not be completed
        '''
        return self.client.get('/v1/shopping/flight-offers', **params)



from amadeus.mixins.parser import Parser


class Response(Parser, object):
    '''
    The response object returned for every API call.

    :var http_response: the raw http response

    :var request: the original Request object used to make this call
    :vartype request: amadeus.Request

    :var result: the parsed JSON received from the API, if the result was JSON
    :vartype result: dict

    :var data: the data extracted from the JSON data, if the body contained
        JSON
    :vartype data: dict

    :var body: the raw body received from the API
    :vartype body: str

    :var parsed: wether the raw body has been parsed into JSON
    :vartype parsed: bool

    :var status_code: The HTTP status code for the response, if any
    :vartype status_code: int
    '''

    # Initialize the Response object with the
    # HTTPResponse object to parse, the client that made the request
    # and the original request made
    def __init__(self, http_response, request):
        self.http_response = http_response
        self.request = request

    # PROTECTED

    # Parses the response, using the client to log any errors
    def _parse(self, client):
        self._parse_status_code()
        self._parse_data(client)
        return self



from .amadeus import Client
from .version import version

from .client.location import Location
from .client.direction import Direction
from .client.request import Request
from .client.response import Response
from .client.errors import ResponseError
from .client.errors import ParserError, ServerError, AuthenticationError
from .client.errors import NotFoundError, ClientError, NetworkError

__all__ = [
    'Client', 'Location', 'Direction', 'version', 'ResponseError',
    'ParserError', 'ServerError', 'AuthenticationError',
    'NotFoundError', 'ClientError', 'Request', 'Response',
    'NetworkError'
]



from amadeus.client.decorator import Decorator


class Airlines(Decorator, object):
    def get(self, **params):
        '''
        Returns the name of the airline given an IATA code.

        .. code-block:: python

            amadeus.reference_data.airlines.get(IATACode='U2')

        :param IATACode: the IATA code for the airline, e.g. ``"1X"``

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: if the request could not be completed
        '''
        return self.client.get('/v1/reference-data/airlines', **params)



from amadeus.client.decorator import Decorator


class HotelOffers(Decorator, object):
    def __init__(self, client, hotel_id):
        Decorator.__init__(self, client)
        self.hotel_id = hotel_id

    def get(self, **params):
        '''
        Get available offers for a specific hotel

        .. code-block:: python

            amadeus.shopping.hotel('SMPARCOL').hotel_offers.get()

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: if the request could not be completed
        '''
        return self.client.get(
            '/v1/shopping/hotels/{0}/hotel-offers'.format(self.hotel_id),
            **params
        )



from platform import python_version
from pprint import pformat
try:
    from urllib.error import URLError
except ImportError:  # pragma: no cover
    from urllib2 import Request as URLError  # pragma: no cover

from amadeus.version import version
from amadeus.client.request import Request
from amadeus.client.response import Response
from amadeus.client.access_token import AccessToken


# A helper module for making generic API calls. It is used by
# every namespaced API method.
class HTTP(object):
    '''
    A helper module for making generic API calls. It is used by
    every namespaced API method.
    '''

    def get(self, path, **params):
        '''
        A helper function for making generic GET requests calls. It is used by
        every namespaced API GET method.

        It can be used to make any generic API call that is automatically
        authenticated using your API credentials:

        .. code-block:: python

            amadeus.get('/foo/bar', airline='1X')

        :param path: path the full path for the API call
        :paramtype path: str

        :param params: (optional) params to pass to the API
        :paramtype params: dict

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: when the request fails
        '''
        return self.request('GET', path, **params)

    def post(self, path, **params):
        '''
        A helper function for making generic POST requests calls. It is used by
        every namespaced API POST method.

        It can be used to make any generic API call that is automatically
        authenticated using your API credentials:

        .. code-block:: python

            amadeus.post('/foo/bar', airline='1X')

        :param path: path the full path for the API call
        :paramtype path: str

        :param params: (optional) params to pass to the API
        :paramtype params: dict

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: when the request fails
        '''
        return self.request('POST', path, **params)

    def request(self, verb, path, **params):
        '''
        A helper function for making generic POST requests calls. It is used by
        every namespaced API method. It can be used to make any generic API
        call that is automatically authenticated using your API credentials:

        .. code-block:: python

          amadeus.request('GET', '/foo/bar', airline='1X')

        :param verb: the HTTP verb to use
        :paramtype verb: str

        :param path: path the full path for the API call
        :paramtype path: str

        :param params: (optional) params to pass to the API
        :paramtype params: dict

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: when the request fails
        '''
        return self._unauthenticated_request(
            verb, path, params,
            self.__access_token()._bearer_token()
        )

    # PROTECTED

    # Builds the URI, the request object, and makes the actual API calls.
    #
    # Used by the AccessToken to fetch a new Bearer Token
    #
    # Passes the response to a Response object for further parsing.
    #
    def _unauthenticated_request(self, verb, path, params, bearer_token=None):
        request = self.__build_request(verb, path, params, bearer_token)
        self.__log(request)
        return self.__execute(request)

    # PRIVATE

    # Builds a HTTP request object that contains all the information about
    # this request
    def __build_request(self, verb, path, params, bearer_token):
        return Request({
            'host': self.host,
            'verb': verb,
            'path': path,
            'params': params,
            'bearer_token': bearer_token,
            'client_version': version,
            'language_version': python_version(),
            'app_id': self.custom_app_id,
            'app_version': self.custom_app_version,
            'ssl': self.ssl,
            'port': self.port
        })

    # Executes the request and wraps it in a Response
    def __execute(self, request):
        http_response = self.__fetch(request)
        response = Response(http_response, request)._parse(self)
        self.__log(response)
        response._detect_error(self)
        return response

    # A memoized AccessToken object, so we don't keep reauthenticating
    def __access_token(self):
        if (not hasattr(self, 'access_token')):
            self.access_token = AccessToken(self)
        return self.access_token

    # Log any object
    def __log(self, object):
        if (self.log_level == 'debug'):
            self.logger.debug(
                '%s\n%s', object.__class__.__name__, pformat(object.__dict__)
            )

    # Actually make the HTTP call, making sure to catch it in case of an error
    def __fetch(self, request):
        try:
            return self.http(request.http_request)
        except URLError as exception:
            return exception



from amadeus.client.decorator import Decorator


class Location(Decorator, object):
    def __init__(self, client, location_id):
        Decorator.__init__(self, client)
        self.location_id = location_id

    def get(self, **params):
        '''
        Returns details for a specific airport.

        .. code-block:: python

            amadeus.reference_data.location('ALHR').get()

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: if the request could not be completed
        '''
        return self.client.get('/v1/reference-data/locations/{0}'
                               .format(self.location_id), **params)



from amadeus.client.decorator import Decorator
from amadeus.reference_data.urls._checkin_links import CheckinLinks


class Urls(Decorator, object):
    def __init__(self, client):
        Decorator.__init__(self, client)
        self.checkin_links = CheckinLinks(client)



from amadeus.client.decorator import Decorator
from amadeus.reference_data.locations._airports import Airports


class Locations(Decorator, object):
    def __init__(self, client):
        Decorator.__init__(self, client)
        self.airports = Airports(client)

    def get(self, **params):
        '''
        Returns details for a specific airport.

        .. code-block:: python


            from amadeus import Location

            client.reference_data.locations.get(
                keyword='lon',
                subType=Location.ANY
            )

        :param keyword: keyword that should represent the  start of
            a word in a city or airport name or code

        :param subType: a comma seperate list of location types to search
            for. You can use :class:`amadeus.Location` as a helper for this.

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: if the request could not be completed
        '''
        return self.client.get('/v1/reference-data/locations', **params)



# Support Python 2 and 3 API calls without importing
# a 3rd party library
try:
    from urllib.request import Request as HTTPRequest
    from urllib.parse import urlencode
except ImportError:  # pragma: no cover
    from urllib2 import Request as HTTPRequest  # pragma: no cover
    from urllib import urlencode  # pragma: no cover


class Request(object):
    '''
    An object containing all the compiled information about the request made.

    :var host: The host used for this API call
    :vartype host: str

    :var port: The port for this API call. Standard set to 443.
    :vartype port: int

    :var ssl: Wether to use SSL for a call, defaults to true
    :vartype ssl: bool

    :var scheme: The scheme used to make the API call
    :vartype scheme: str

    :var params: The GET/POST params for the API call
    :vartype params: dict

    :var path: The path of the API to be called
    :vartype path: str

    :var verb: The verb used to make an API call ('GET' or 'POST')
    :vartype verb: str

    :var bearer_token: The bearer token (if any) that was used for
        authentication
    :vartype bearer_token: str

    :var headers: The headers used for the API call
    :vartype headers: dict

    :var client_version: The library version used for this request
    :vartype client_version: str

    :var language_version: The Python language version used for this request
    :vartype language_version: str

    :var app_id: The custom app ID passed in for this request
    :vartype app_id: str

    :var app_version: The custom app version used for this request
    :vartype app_version: str
    '''

    def __init__(self, options):
        self.http_request = None
        self.__initialize_options(options)
        self.__initialize_headers()
        self.__build_http_request()

    # PRIVATE

    # Initializes the Request object with all the passed in params. These
    # do not change once set
    def __initialize_options(self, options):
        self.host = options['host']
        self.port = options['port']
        self.ssl = options['ssl']
        self.scheme = 'https' if self.ssl else 'http'
        self.verb = options['verb']
        self.path = options['path']
        self.params = options['params']
        self.bearer_token = options['bearer_token']
        self.client_version = options['client_version']
        self.language_version = options['language_version']
        self.app_id = options['app_id']
        self.app_version = options['app_version']

    # Initializes the basic headers
    def __initialize_headers(self):
        self.headers = {
            'User-Agent': self.__build_user_agent(),
            'Accept': 'application/json, application/vnd.amadeus+json'
        }

    # Determines the User Agent
    def __build_user_agent(self):
        user_agent = 'amadeus-python/{0}'.format(self.client_version)
        user_agent += ' python/{0}'.format(self.language_version)
        if self.app_id:
            user_agent += ' {0}/{1}'.format(self.app_id, self.app_version)
        return user_agent

    # Builds up a HTTP Request objectm if not akready set
    def __build_http_request(self):
        request = self.__request_for_verb()
        self.__add_post_data_header()
        self.__add_bearer_token_header()
        self.__apply_headers(request)
        self.http_request = request

    # Builds a HTTP Request object based on the path, params, and verb
    def __request_for_verb(self):
        params = self._encoded_params().encode()
        path = self.__full_url()

        if (self.verb == 'GET'):
            return HTTPRequest(path)
        else:
            return HTTPRequest(path, params)

    # Encodes the params before sending them
    def _encoded_params(self):
        return urlencode(self.params)

    # Builds up the full URL based on the scheme, host, path, and params
    def __full_url(self):
        full_url = '{0}://{1}'.format(self.scheme, self.host)
        if not self.__port_matches_scheme():
            full_url = '{0}:{1}'.format(full_url, self.port)
        full_url = '{0}{1}'.format(full_url, self.path)
        if (self.verb == 'GET'):
            full_url += '?{0}'.format(self._encoded_params())
        return full_url

    def __port_matches_scheme(self):
        return ((self.ssl and self.port == 443) or
                (not self.ssl and self.port == 80))

    # Adds an extra header if the verb is POST
    def __add_post_data_header(self):
        if (self.verb == 'POST'):
            self.headers['Content-Type'] = 'application/x-www-form-urlencoded'

    # Adds the authentication header if the bearer token has been set
    def __add_bearer_token_header(self):
        if (self.bearer_token is not None):
            self.headers['Authorization'] = self.bearer_token

    # Applies all the headers to the HTTP Request object
    def __apply_headers(self, http_request):
        http_request.headers = self.headers



from pprint import pformat


class ResponseError(RuntimeError):
    '''
    An Amadeus error

    :var response: The response object containing the raw HTTP response and
        the request used to make the API call.
    :vartype response: amadeus.Response

    :var code: A unique code for this type of error. Options include
        ``NetworkError``, ``ParserError``, ``ServerError``,
        ``AuthenticationError``, ``NotFoundError`` and ``UnknownError``.
    :vartype code: str
    '''

    def __init__(self, response):
        self.response = response
        self.code = self.__determine_code()
        RuntimeError.__init__(self, self.description())

    # PROTECTED

    # Log the error
    def _log(self, client):
        if (client.log_level == 'warn'):
            client.logger.warning(
                'Amadeus %s: %s', self.code, pformat(self.description)
            )

    # PRIVATE

    # Determines the description for this error, as used in in the error output
    def description(self):
        description = self.short_description(self.response)
        return description + self.long_description(self.response)

    # Determines the short description, printed after on the same line as the
    # error class name
    def short_description(self, response):
        if hasattr(response, 'status_code') and response.status_code:
            return '[{0}]'.format(response.status_code)
        else:
            return '[---]'

    # Determines the longer description, printed after the initial error
    def long_description(self, response):
        message = ''
        if not(response and response.parsed):
            return message
        if 'error_description' in response.result:
            message += self.error_description(self.response)
        if 'errors' in response.result:
            message += self.errors_descriptions(self.response)
        return message

    # Returns the description of a single error
    def error_description(self, response):
        message = ''
        if 'error' in response.result:
            message += '\n{0}'.format(response.result['error'])
        message += '\n{0}'.format(response.result['error_description'])
        return message

    # Returns the description of multiple errors
    def errors_descriptions(self, response):
        messages = map(self.errors_description, response.result['errors'])
        return ''.join(messages)

    # Returns the description of a single error in a multi error response
    def errors_description(self, error):
        message = '\n'
        if ('source' in error) and ('parameter' in error['source']):
            message += '[{0}] '.format(error['source']['parameter'])
            if 'detail' in error:
                message += error['detail']
        return message

    # sets the error code to the name of this class
    def __determine_code(self):
        return self.__class__.__name__


class NetworkError(ResponseError):
    '''
    This error occurs when there is some kind of error in the network
    '''
    pass


class ParserError(ResponseError):
    '''
    This error occurs when the response type was JSOn but could not be parsed
    '''
    pass


class ServerError(ResponseError):
    '''
    This error occurs when there is an error on the server
    '''
    pass


class ClientError(ResponseError):
    '''
    This error occurs when the client did not provide the right parameters
    '''
    pass


class AuthenticationError(ResponseError):
    '''
    This error occurs when the client did not provide the right credentials
    '''
    pass


class NotFoundError(ResponseError):
    '''
    This error occurs when the path could not be found
    '''
    pass



from amadeus.client.decorator import Decorator


class Airports(Decorator, object):
    def get(self, **params):
        '''
        Returns a list of relevant airports near to a given point.

        .. code-block:: python


            client.reference_data.locations.get(
                longitude=49.0000,
                latitude=2.55
            )

        :param latitude: latitude of geographic location to search around.
            For example: ``52.5238``
        :param longitude: longitude of geographic location to search around.
            For example: ``13.3835``

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: if the request could not be completed
        '''
        return self.client.get(
            '/v1/reference-data/locations/airports', **params)



from amadeus.client.decorator import Decorator

from .hotel._hotel_offers import HotelOffers
from .hotel._offer import Offer


class Hotel(Decorator, object):
    def __init__(self, client, hotel_id):
        Decorator.__init__(self, client)
        self.hotel_id = hotel_id
        self.hotel_offers = HotelOffers(client, hotel_id)

    def offer(self, offer_id):
        return Offer(self.client, self.hotel_id, offer_id)



import os
import sys
import logging

try:
    from urllib.request import urlopen
except ImportError:  # pragma: no cover
    from urllib2 import urlopen  # pragma: no cover


# A set of helper methods to allow the validating of
# arguments past into the Client
class Validator(object):

    # PROTECTED

    # Checks a list of options for unrecognized keys and warns the user.
    # This is mainly used to provide a nice experience when users make a typo
    # in their arguments.
    def _warn_on_unrecognized_options(self, options, logger, valid_options):
        for key in options:
            if (key in valid_options):
                continue
            logger.warning('Unrecognized option: {0}'.format(key))

    # Initializes the credentials, requiring an ID and Secret
    def _initialize_client_credentials(self, options):
        self.client_id = self.__init_required('client_id', options)
        self.client_secret = self.__init_required('client_secret', options)

    # Initializes an optional Logger
    def _initialize_logger(self, options):
        default_logger = logging.getLogger('amadeus')
        default_logger.setLevel(logging.DEBUG)
        handler = logging.StreamHandler(sys.stdout)
        default_logger.addHandler(handler)

        self.logger = self.__init_optional('logger', options, default_logger)
        self.log_level = self.__init_optional('log_level', options, 'silent')

    # Initializes an optional host, hostname, port, and SSL requirements
    def _initialize_host(self, options):
        self.hostname = self.__init_optional('hostname', options, 'test')
        self.host = self.__init_optional('host', options,
                                         self.HOSTS[self.hostname])
        self.ssl = self.__init_optional('ssl', options, True)
        self.port = self.__init_optional('port', options, 443)

    # Initializes an optional custom App ID and Version
    def _initialize_custom_app(self, options):
        self.custom_app_id = self.__init_optional('custom_app_id',
                                                  options, None)
        self.custom_app_version = self.__init_optional('custom_app_version',
                                                       options, None)

    # Initializes a custom http handler
    def _initialize_http(self, options):
        self.http = self.__init_optional('http', options, urlopen)

    # PRIVATE

    # Uses 'init_optional' to find an entry, and it that returns
    # nil it raises an ArgumentError
    #
    def __init_required(self, key, options):
        value = self.__init_optional(key, options)
        if (value is None):
            raise ValueError('Missing required argument: {0}'.format(key))
        return value

    # Tries to find an option by string or symbol in the options hash and
    # in the environment variables.When it can not find it anywhere it
    # defaults to the provided default option.
    def __init_optional(self, key, options, defa_ult=None):
        value = options.get(key, None)
        if (value is None):
            env_key = 'AMADEUS_{0}'.format(key.upper())
            value = os.environ.get(env_key, None)
        if (value is None):
            value = defa_ult
        return value



import json

from amadeus.client.errors import ParserError, ServerError
from amadeus.client.errors import AuthenticationError, NetworkError
from amadeus.client.errors import NotFoundError, ClientError


class Parser(object):

    # PROTECTED

    # Tries to detect for appropriate errors
    def _detect_error(self, client):
        error = self.error_for(self.status_code, self.parsed)
        if error is not None:
            self.__raise_error(error, client)

    def error_for(self, status_code, parsed):  # noqa: C901
        if status_code is None:
            return NetworkError
        if status_code >= 500:
            return ServerError
        if status_code == 401:
            return AuthenticationError
        if status_code == 404:
            return NotFoundError
        if status_code >= 400:
            return ClientError
        if not parsed:
            return ParserError

    # Parses the HTTP status code
    def _parse_status_code(self):
        http_response = self.http_response
        self.status_code = getattr(http_response, 'status', None)
        self.status_code = getattr(http_response, 'code', self.status_code)

    # Tries to parse the received data from raw string to parsed data and into
    # a data object
    def _parse_data(self, client):
        self.parsed = False
        self.data = None
        self.body = None
        self.result = None
        self.headers = {}

        self.__parse_body_and_headers(self.http_response, client)
        self.result = self.__parse_json(client)
        if (self.result is not None):
            self.data = self.result.get('data', None)

    # PRIVATE

    # Logs and raises the error
    def __raise_error(self, error_class, client):
        error = error_class(self)
        error._log(client)
        raise error

    # Extract the body and headers
    def __parse_body_and_headers(self, http_response, client):
        if hasattr(http_response, 'getheaders'):
            self.headers = dict(http_response.getheaders()) or self.headers
        if hasattr(http_response, 'info'):
            self.headers = http_response.info() or self.headers
        if hasattr(http_response, 'read'):
            self.body = http_response.read()
        if hasattr(self.body, 'decode'):
            self.body = self.body.decode('utf8')

    # Tries to parse the JSON, if there is any
    def __parse_json(self, client):
        try:
            if (self.__is_json()):
                result = json.loads(self.body)
                self.parsed = True
                return result
            else:
                return None
        except Exception:
            self.__raise_error(ParserError, client)

    # checks if the HTTPResponse included JSON
    def __is_json(self):
        return self.__has_json_header() and self.__has_body()

    # checks if the HTTPResponse has a non-empty body
    def __has_body(self):
        return self.body and len(self.body) > 0

    # checks if the HTTPResponse has a JSON header
    def __has_json_header(self):
        content_type = self.headers.get('Content-Type', None)
        if (content_type is not None):
            types = content_type.split(';')[0]
            types = ['application/json', 'application/vnd.amadeus+json']
            return content_type in types
        else:
            return False



from amadeus.client.decorator import Decorator


class HotelOffers(Decorator, object):
    def get(self, **params):
        '''
        Search for hotels and retrieve availability and rates information.

        Use either `cityCode`, `longitude` and `latitude` or `hotels` to
        make a search.

        .. code-block:: python

            amadeus.shopping.hotel_offers.get(cityCode='PAR')
            amadeus.shopping.hotel_offers.get(longitude=49.0, latitude=2.0)
            amadeus.shopping.hotel_offers.get(hotels='RTPAR001, RTPAR002')

        :param cityCode: the City IATA code for which to find a hotel, for
            example ``"BOS"`` for Boston.
        :param latitude: latitude of geographic location to search around.
            For example: ``52.5238``
        :param longitude: longitude of geographic location to search around.
            For example: ``13.3835``
        :param hotels: Comma separated list of Amadeus hotel codes to request.
            Example: ``RTPAR001``

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: if the request could not be completed
        '''
        return self.client.get('/v1/shopping/hotel-offers', **params)



from amadeus.client.decorator import Decorator
from amadeus.reference_data._urls import Urls
from amadeus.reference_data._location import Location
from amadeus.reference_data._locations import Locations
from amadeus.reference_data._airlines import Airlines


class ReferenceData(Decorator, object):
    def __init__(self, client):
        Decorator.__init__(self, client)
        self.urls = Urls(client)
        self.locations = Locations(client)
        self.airlines = Airlines(client)

    def location(self, location_id):
        return Location(self.client, location_id)



from amadeus.client.decorator import Decorator
from amadeus.travel._analytics import Analytics


class Travel(Decorator, object):
    def __init__(self, client):
        Decorator.__init__(self, client)
        self.analytics = Analytics(client)



#
class Location(object):
    '''
    A list of location types, as used in searching for locations

    .. code-block:: python


        from amadeus import Location

        client.reference_data.locations.get(
            keyword='lon',
            subType=Location.ANY
        )

    :cvar AIRPORT: ``"AIRPORT"``
    :cvar CITY: ``"CITY"``
    :cvar ANY: ``"AIRPORT,CITY"``
    '''
    # Airport
    AIRPORT = 'AIRPORT'
    # City
    CITY = 'CITY'
    # Any
    ANY = ','.join([AIRPORT, CITY])



#
class Direction (object):
    '''
    A list of direction types, as used in Busiest Travel Period

    .. code-block:: python


        from amadeus import Direction

        client.travel.analytics.air_traffic.busiest_period.get(
            cityCode = 'MAD',
            period = '2017',
            direction = Direction.ARRIVING
        )

    :cvar ARRIVING: ``"ARRIVING"``
    :cvar DEPARTING: ``"DEPARTING"``
    '''
    # Arriving
    ARRIVING = 'ARRIVING'
    # Departing
    DEPARTING = 'DEPARTING'



from .mixins.validator import Validator
from .mixins.http import HTTP
from .mixins.pagination import Pagination
from .namespaces import Core as Namespaces


class Client(Namespaces, Pagination, Validator, HTTP, object):
    '''
    The Amadeus client library for accessing
    the travel APIs.
    '''

    # The available hosts for this API
    HOSTS = {
        'test': 'test.api.amadeus.com',
        'production': 'api.amadeus.com'
    }

    # The initialization method for the entire module
    def __init__(self, **options):
        '''
        Initialize using your credentials:

        .. code-block:: python


            from amadeus import Client

            amadeus = Client(
                client_id='YOUR_CLIENT_ID',
                client_secret='YOUR_CLIENT_SECRET'
            )

        Alternatively, initialize the library using the environment variables
        ``AMADEUS_CLIENT_ID`` and ``AMADEUS_CLIENT_SECRET``.

        .. code-block:: python


              amadeus = amadeus.Client()

        :param client_id: the API key used to authenticate the API
        :paramtype client_id: str

        :param client_secret: the API secret used to authenticate the API
        :paramtype client_secret: str

        :param logger: (optional) a Python compatible logger
            (Default: ``logging.Logger``)
        :paramtype logger: logging.Logger

        :param log_level: (optional) the log level of the client, either
            ``"debug"``, ``"warn"``, or ``"silent"`` mode
            (Default: ``"silent"``)
        :paramtype log_level: str

        :param hostname: (optional) the name of the server API calls are made
            to, ``"production"``  or ``"test"``. (Default: ``"test"``)
        :paramtype hostname: str

        :param host: (optional) alternatively, you can specify a full host
            domain name instead, e.g. ``"api.example.com"``
        :paramtype host: str

        :param ssl: if this client is should use HTTPS (Default: ``True``)
        :paramtype ssl: bool

        :param port: the port this client should use (Default: ``80`` for HTTP
            and ``443`` for HTTPS)
        :paramtype port: int

        :param custom_app_id: (optional) a custom App ID to be passed in
            the User Agent to the server (Default: ``None``)
        :paramtype custom_app_id: str

        :param custom_app_version: (optional) a custom App Version number to
            be passed in the User Agent to the server (Default: ``None``)
        :paramtype custom_app_version: str

        :param http: (optional) a :func:`urllib.request.urlopen` compatible
            client that accepts a :class:`urllib.request.Request` compatible
            object (Default: ``urlopen``)
        :paramtype http: urllib.request.urlopen

        :raises ValueError: when a required param is missing
        '''
        self._initialize_client_credentials(options)
        self._initialize_logger(options)
        self._initialize_host(options)
        self._initialize_custom_app(options)
        self._initialize_http(options)

        recognized_options = ['client_id', 'client_secret', 'logger', 'host',
                              'hostname', 'custom_app_id',
                              'custom_app_version', 'http',
                              'log_level', 'ssl', 'port']
        self._warn_on_unrecognized_options(options, self.logger,
                                           recognized_options)
        Namespaces.__init__(self)



from amadeus.client.decorator import Decorator


class CheckinLinks(Decorator, object):
    def get(self, **params):
        '''
        Returns the checkin links for an airline, for the
        language of your choice

        .. code-block:: python

            amadeus.reference_data.urls.checkin_links.get(airline='1X')

        :param airline: the IATA code for the airline, e.g. ``"1X"``
        :param language: the locale for the links, for example ``"en-GB"``

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: if the request could not be completed
        '''
        return self.client.get(
            '/v2/reference-data/urls/checkin-links', **params)



from amadeus.client.decorator import Decorator
from amadeus.shopping._flight_dates import FlightDates
from amadeus.shopping._flight_destinations import FlightDestinations
from amadeus.shopping._flight_offers import FlightOffers
from amadeus.shopping._hotel_offers import HotelOffers
from amadeus.shopping._hotel import Hotel


class Shopping(Decorator, object):
    def __init__(self, client):
        Decorator.__init__(self, client)
        self.flight_dates = FlightDates(client)
        self.flight_destinations = FlightDestinations(client)
        self.flight_offers = FlightOffers(client)
        self.hotel_offers = HotelOffers(client)

    def hotel(self, hotel_id):
        return Hotel(self.client, hotel_id)


__all__ = ['FlightDates', 'FlightDestinations', 'FlightOffers']



from amadeus.client.decorator import Decorator


class BusiestPeriod(Decorator, object):
    def get(self, **params):
        '''
        Returns a list of air traffic reports, based on number of travelers.

        .. code-block:: python

            amadeus.travel.analytics.air_traffic.busiest_period.get(
                origin='MAD',
                period='2017',
                direction=Direction.ARRIVING
            )

        :param cityCode: IATA code of the origin city, for
            example ``"BOS"`` for Boston.
        :param query: period when consumers are traveling
            in ``YYYY`` format

        :rtype: amadeus.Response
        :raises amadeus.ResponseError: if the request could not be completed
        '''
        return self.client.get('/v1/travel/analytics/air-traffic/busiest-period',
                               **params)



import io
import os

from setuptools import setup, find_packages

# Load the version number
about = {}
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'amadeus', 'version.py')) as f:
    exec(f.read(), about)

# Import the README and use it as the long-description.
with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

setup(
    name='amadeus',
    version=about['version'],
    description='Python module for the Amadeus travel APIs',
    long_description=long_description,
    author='Amadeus',
    author_email='developers@amadeus.com',
    python_requires='>=2.7.0',
    url='https://github.com/amadeus4dev/amadeus-python',
    install_requires=[],
    packages=find_packages(),
    data_files=[('docs', ['README.rst', 'CHANGELOG.rst'])],
    include_package_data=True,
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



import copy


# A set of helper methods to allow the validating of
# arguments past into the Client
class Pagination(object):

    def previous(self, response):
        return self.__page('previous', response)

    def next(self, response):
        return self.__page('next', response)

    def first(self, response):
        return self.__page('first', response)

    def last(self, response):
        return self.__page('last', response)

    # PRIVATE

    def __page(self, name, response):
        page_number = self.__page_number_for(name, response)
        if page_number is None:
            return None
        params = copy.deepcopy(response.request.params)
        if 'page' not in params:
            params['page'] = {}
        params['page']['offset'] = page_number
        return self.request(
            response.request.verb,
            response.request.path,
            **params
        )

    def __page_number_for(self, name, response):
        try:
            return response.result['meta']['links'][name].split('=')[-1]
        except Exception:
            return None
